(* Graph, a terminal time series graphing utility
   Copyright (C) 2007 Eric Stokes, N. Owen Gunden

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. *)

open Unix
open ANSITerminal
open Args

let pad_or_trim ~to_length s =
  if String.length s >= to_length then String.sub s 0 to_length else
  let num_zeros = to_length - String.length s in
  String.make num_zeros '0' ^ s

let draw_y_axis td c =
  let width = c.y_axis_width in
  let height = c.range_high -. c.range_low in
  for row = 0 to c.rows - 1 do
    let v = 
      let v = (float (c.rows - row - 1)) *. height /. (float c.rows) +. c.range_low in
      match c.scale with
      | `Log base -> base ** v
      | `Linear -> v
    in
    let s =
      if c.range_int then
        (pad_or_trim ~to_length:width (string_of_int (Pervasives.truncate v)))
      else
        (pad_or_trim ~to_length:width (Printf.sprintf "%*.2f" width v))
    in
    Termdraw.write_string td ~row ~col:0 [Foreground Blue] s;
  done

exception Break

(* fills in a column from floor up to height *)
let column td c ~col height =
  let color height row =
    let v = (float (c.rows - row)) *. c.scale_factor in
    if (float height) *. c.scale_factor < c.green_range_low
    then Red
    else if v < c.green_range_high
    then Green
    else Red
  in
  match height with
  | Some height when height = 0 ->
      Termdraw.set_pixel td '=' [Background (color 0 0)] ~row:(c.rows - 1) ~col
  | Some height when height < 0 ->
      Termdraw.set_pixel td '-' [Background (color 0 0)] ~row:(c.rows - 1) ~col
  | Some height ->
    begin
      (* baseline green *)
      Termdraw.set_pixel td '.' 
        [Background (color height (c.rows - 2)); Hidden] 
        ~col ~row:(c.rows - 1);
      (* rest of the column is red or green *)
      try
        for row = c.rows - 2 downto (c.rows - height - 1) do
          if row < 1 then begin
            let chr = if height = c.rows - 1 then '=' else '+' in
            Termdraw.set_pixel td chr [Background Yellow] ~col ~row:0;
            raise Break
          end;
          Termdraw.set_pixel td '.' [Background (color height (row + 1)); Hidden] ~col ~row
        done
      with Break -> ()
    end
  | None -> ()

let normalize_sample c sample =
  let sample = 
    match c.scale with
    | `Log base -> (log sample) /. (log base)
    | `Linear -> sample
  in
  Pervasives.truncate (sample /. c.scale_factor)

let draw_sample ~q ~td ~c ~title sample = 
  try
    let normalized_sample = normalize_sample c sample in
    ArrayQueue.push q normalized_sample;
    Termdraw.fill td '.' [Hidden];
    draw_y_axis td c;
    for col = c.y_axis_width to c.cols - 1 do
      column td c ~col (ArrayQueue.get q (col + 1))
    done;
    Termdraw.write_string td ~row:0 ~col:(c.y_axis_width + 1)
      [Foreground Blue; Background White] title;
    Termdraw.write_string td ~row:1 ~col:(c.y_axis_width + 1) [Foreground Green]
      (string_of_float sample);

    Termdraw.redraw td
  with exn -> 
    Printf.eprintf "fatal error: %s\n" (Printexc.to_string exn);
    exit 1

let get_data =
  let rex = Pcre.regexp "\\s+" in
  let lastdraw = ref 0. in
  let samples = ref [] in
  let add_sample sample = samples := sample :: !samples in
  let avg_sample () =
    match !samples with
    | [] -> assert false
    | samples ->
        List.fold_left (+.) 0. samples /. float (List.length samples)
  in
  (fun c line ->
    let this_sample =
      let s = match c.field with
      | `All -> line
      | `Specific i -> 
          let tokens = Pcre.split ~rex line in
          List.nth tokens i
      in float_of_string s
    in
    add_sample this_sample;
    let now = Unix.gettimeofday () in
    if now > !lastdraw +. c.mindelay then
      begin
        let res = avg_sample () in
        samples := [];
        lastdraw := now;
        Some res
      end
    else
      None
  )

let () =
  let c = Args.parse () in
  if c.y_axis_width >= c.cols then
    failwith "y_axis_width must be less than COLUMNS";
  let td = 
    Termdraw.create 
      ~rows:c.rows ~cols:c.cols 
      ~init_styles:[Hidden] ~init_character:'.' 
  in
  Termdraw.redraw td;
  let q = ArrayQueue.create (c.cols + 1) in
  while true do
    let line = input_line Pervasives.stdin in
    try
      match get_data c line with
      | None -> ()
      | Some data -> draw_sample ~title:c.title ~q ~td ~c data
    with exn -> 
      Printf.eprintf 
        "invalid sample: %s, %s\n%!" 
        line (Printexc.to_string exn)
  done
